#include "plan_manager.h"
#include "Events/EventRouter.h"
#include "Events/TextMsgEvent.h"
#include "Motion/MotionManager.h"
#include "Motion/MotionSequenceMC.h"
#include "Sound/SoundManager.h"
#include <string>
#include "actions/locate_ball.h"
#include "actions/colocate_to_ball.h"
#include "actions/check_room.h"
#include "actions/check_holding.h"
#include "actions/locate_dest.h"
#include "actions/grab_ball.h"
#include "actions/go_other_room.h"
#include "actions/release_ball.h"
#include "utils/effect_checker.h"

//#define CHECKING_EFFECTS

PlanManager * manager = NULL;

static const char locateBallStr[]	= "lb";
static const char checkRoomStr[]	= "cr";
static const char locateDestStr[]	= "ld";
static const char checkHoldStr[]	= "ch";
static const char colocateStr[]		= "co";
static const char grabBallStr[]		= "gb";
static const char goRoomStr[]		= "go";
static const char releaseBallStr[]	= "rb";
static const char stopStr[]			= "stop";
static const char printCoordsStr[]	= "pcoords";
static const char printStateStr[]	= "pstate";
static const char downStr[]			= "down";
static const char doneStr[]			= "done";
static const char replanStr[]		= "rp";

#ifdef CHECKING_EFFECTS
	static EffectChecker* effect_check = NULL;
#endif

PlanManager::PlanManager() : BehaviorBase("PlanManager"),
/*	ballX(0),
	ballY(0),
	destX(0),
	destY(0),
	robotA(0),*/
	destDist(-1),
	destBearing(-1),
	inGreenRm(UNKNOWN),
	ballInGreenRm(UNKNOWN),
	holdingBall(UNKNOWN),
	actionFailed(false),
	cmdsock(NULL),
	doingCheck(false)
	//lastMot(NULL)
	{
	manager = this;

	// Set up the actions
	actionArray[0] = new LocateBall();
	actionArray[1] = new CheckRoom();
	actionArray[2] = new LocateDest();
	actionArray[3] = new CheckHolding();
	actionArray[4] = new ColocateToBall();
	actionArray[5] = new GrabBall();
	actionArray[6] = new GoOtherRm();
	actionArray[7] = new ReleaseBall();
	
	sndman->LoadFile("replan.wav");
}

PlanManager::~PlanManager() {
	manager = NULL;
	//delete lastMot;
	// Deactivate any active states and remove them
	for(int i=0 ; i < 8 ; i++) 
		if(actionArray[i] != NULL) {
			if(actionArray[i]->isActive()) actionArray[i]->DoStop();
			delete actionArray[i];
		}
	sndman->ReleaseFile("replan.wav");
}

void PlanManager::DoStart() {
	BehaviorBase::DoStart(); // do this first (required)
	
	// Before we do anything, let's get up off the floor
	SharedObject<MediumMotionSequenceMC> mseq_mc;
	mseq_mc->LoadFile("get_up.mot");
	motman->addPrunableMotion(mseq_mc);
	
	//erouter->addListener(this,EventBase::locomotionEGID); //,EventBase::statusETID
	erouter->addListener(this,EventBase::aiEGID);
	erouter->addListener(this,EventBase::textmsgEGID);
	
	// Turn on wireless
	cmdsock=wireless->socket(SocketNS::SOCK_STREAM, 300, 300);
	wireless->setDaemon(cmdsock,true);
	wireless->setReceiver(cmdsock->sock, aiComm);
	wireless->listen(cmdsock->sock, 10101);
}

void PlanManager::DoStop() {
	lieDown();
	// Close socket; turn wireless off
	wireless->setDaemon(cmdsock,false);
	wireless->close(cmdsock);
	BehaviorBase::DoStop(); // do this last (required)
}

void PlanManager::processEvent(const EventBase& event) {
cout << "Manager::processEvent " << event.getName() << endl;
	
	// Note, we can't use a switch here because gcc doesn't like
	// strings inside of switches.
	/*if(event.getGeneratorID() == EventBase::locomotionEGID) {
		//updateMap(reinterpret_cast<const LocomotionEvent*>(&event));	
	}*/
	if(event.getGeneratorID() == EventBase::textmsgEGID) {
		const TextMsgEvent &txtev = *reinterpret_cast<const TextMsgEvent*>(&event);
		const std::string msgText = txtev.getText();
		runCommand(msgText);
	}
	// Action ended, communicate back to LISP
	else if(event.getGeneratorID() == EventBase::aiEGID) {
		std::cout << "Manager: aiEvent, " << printActionName(event.getSourceID()) << std::endl;

#ifdef CHECKING_EFFECTS 
		if(event.getSourceID() == EFFECT_CHECK) {
			if(effect_check->needReplan()) {
				std::cout << "Requesting replan" << std::endl;
				cmdsock->printf(effect_check->getReplanString());
				sndman->PlayFile("replan.wav");
			}
			else cmdsock->printf("T");
			effect_check->DoStop();
			doingCheck = false;
			return;
		}
		if(doingCheck) {
			return; // Let the effect_checker take care of it
		}
		else if(actionFailed) { // The action failed
			cmdsock->printf("RUUN "); // So replan
		}
#else
		if(actionFailed) { // The action failed
			cmdsock->printf("F"); // So replan
		}
#endif
		else {
			switch(event.getSourceID()) {
			case LOCATE_BALL:
				if(ballInGreenRm == YES) cmdsock->printf("Y");
				else if (ballInGreenRm == NO) cmdsock->printf("N");
				else cmdsock->printf("U");
				break;
			case CHECK_ROOM:
				if(inGreenRm == YES) cmdsock->printf("Y");
				else if(inGreenRm == NO) cmdsock->printf("N");
				else cmdsock->printf("U");
				break;
			case CHECK_HOLDING:
				if(holdingBall == YES) cmdsock->printf("Y");
				else if(holdingBall == NO) cmdsock->printf("N");
				else cmdsock->printf("U");
				break;
#ifdef CHECKING_EFFECTS			
			case GRAB_BALL:
				doingCheck = true;
				// For the purpose of checking effects we really only care
				// about whether we're holding the ball
				effect_check = new EffectChecker(UNKNOWN, UNKNOWN, YES);
				effect_check->DoStart();
				break;
			case GO_OTHER_RM:
				doingCheck = true;
				// Make sure everything is as we think it is
				effect_check = new EffectChecker(UNKNOWN, UNKNOWN, holdingBall);
				effect_check->DoStart();
				break;
			case RELEASE_BALL:
				doingCheck = true;
				// For the purpose of checking effects we really only care
				// about whether we're holding the ball
				effect_check = new EffectChecker(UNKNOWN, UNKNOWN, NO);
				effect_check->DoStart();
				break;
#endif
			default:
				cmdsock->printf("T");
			}
		}
	}
	else {
			ASSERT(false,"Manager: unprocessed event " << event.getName() << endl);
	}
}

// Proccess command strings
void PlanManager::runCommand(const std::string& msgText) {
	actionFailed = false; // clear flag
	// Again, we can't use a switch here because of the gcc bug
	std::cout << "Manager: command input, " << msgText << std::endl;

	if(msgText.compare(locateBallStr) ==  0) { // Locate ball command
		actionArray[LOCATE_BALL]->DoStart();
	}
	else if(msgText.compare(checkRoomStr) == 0) {
		actionArray[CHECK_ROOM]->DoStart();
	}
	else if(msgText.compare(locateDestStr) == 0) {
		actionArray[LOCATE_DEST]->DoStart();
	}
	else if(msgText.compare(checkHoldStr) == 0) {
		actionArray[CHECK_HOLDING]->DoStart();
	}
	else if(msgText.compare(colocateStr) == 0) {
		actionArray[COLOCATE]->DoStart();
	}
	else if(msgText.compare(grabBallStr) == 0) {
		actionArray[GRAB_BALL]->DoStart();
	}
	else if(msgText.compare(goRoomStr) == 0) {
		actionArray[GO_OTHER_RM]->DoStart();
	}
	else if(msgText.compare(releaseBallStr) == 0) {
		actionArray[RELEASE_BALL]->DoStart();
	}
	else if(msgText.compare(stopStr) == 0) {
		// Stop any running action
		for(int i=0 ; i < 8 ; i++)
			if(actionArray[i] != NULL && actionArray[i]->isActive())
				actionArray[i]->DoStop();
	}
	/*else if(msgText.compare(printCoordsStr) == 0) {
		std::cout << "Current coodinate information:"
		          << "\n\tBall: x=" << ballX << "  y=" << ballY
		          << "\n\tDest: x=" << destX << "  y=" << destY
		          << std::endl;
	}*/
	else if(msgText.compare(printStateStr) == 0) {
		printState();
	}
	else if(msgText.compare(downStr) == 0) {
		lieDown();
	}
	else if(msgText.compare(replanStr) == 0) {
		cmdsock->printf("T");
		sndman->PlayFile("replan.wav");
	}
	else if(msgText.compare(doneStr) == 0) {
		sndman->LoadFile("finished.wav");
		//SharedObject<MediumMotionSequenceMC> mseq_mc;
		//mseq_mc->LoadFile("vicdance.mot");
		//motman->addPrunableMotion(mseq_mc);
		cmdsock->printf("T");
		sndman->PlayFile("finished.wav");
		sndman->ReleaseFile("finished.wav");
	}
	else ASSERT(false,"Manager: Invalid command." << endl);
}

void PlanManager::printState() {
	std::cout << "inGreenRoom: ";
	switch(inGreenRm) {
	case YES:
		std::cout << "YES" << endl;
		break;
	case NO:
		std::cout << "NO" << endl;
		break;
	case UNKNOWN:
		std::cout << "UNKNOWN" << endl;
		break;
	default:
		std::cout << "!Bad value!" << endl;
	}
	
	std::cout << "ballInGreenRoom: ";
	switch(ballInGreenRm) {
	case YES:
		std::cout << "YES" << endl;
		break;
	case NO:
		std::cout << "NO" << endl;
		break;
	case UNKNOWN:
		std::cout << "UNKNOWN" << endl;
		break;
	default:
		std::cout << "!Bad value!" << endl;
	}
	
	std::cout << "holdingBall: ";
	switch(holdingBall) {
	case YES:
		std::cout << "YES" << endl;
		break;
	case NO:
		std::cout << "NO" << endl;
		break;
	case UNKNOWN:
		std::cout << "UNKNOWN" << endl;
		break;
	default:
		std::cout << "!Bad value!" << endl;
	}
}

/*void PlanManager::updateMap(const LocomotionEvent* lEvent) {
std::cout << "Last: " << lastMot->getDescription(true,3) << std::endl;
std::cout << "Current: " << lEvent->getDescription(true,3) << std::endl;
	// Calculate the egocentric changes
	static float dx,dy,da;
	static double dt;
	dt = (lEvent->getTimeStamp() - lastMot->getTimeStamp());///100000;
	dx = (lastMot->x)*dt;
	dy = (lastMot->y)*dt;
	da = (lastMot->a)*dt;

std::cout << "dt=" << dt << "\tdx=" << dx << "\tdy=" << dy << "\tda=" << da << std::endl;

	// And calculate the offset coordinate changes in x and y to apply
	// to the location of the ball and destination
	static float dxo, dyo;
	dxo = cos(da)*dx - sin(da)*dy;
	dyo = sin(da)*dx + cos(da)*dy;

std::cout << "dxo=" << dxo << "\tdyo=" << dyo  << std::endl;
		
	// Update the important location information
	ballX += dxo;
	ballY += dyo;
	destX += dxo;
	destY += dyo;
	robotA += da;
	
	// Save a copy of most recent event for future reference.
	delete lastMot;
	lastMot = reinterpret_cast<const LocomotionEvent*>(lEvent->clone());
}*/

void PlanManager::lieDown() {
	SharedObject<MediumMotionSequenceMC> mseq_mc;
	mseq_mc->advanceTime(1000);
	mseq_mc->LoadFile("walk.pos");
	mseq_mc->advanceTime(2000);
	mseq_mc->LoadFile("hl_close.pos");
	mseq_mc->advanceTime(5000);
	mseq_mc->LoadFile("liedown.pos");
	motman->addPrunableMotion(mseq_mc);
	cmdsock->printf("T");
}

// The command packet reassembly mechanism
int PlanManager::aiComm(char *buf, int bytes) {
	if(manager==NULL)
		return 0;
	static std::string cmd;
	for(int i=0; i<bytes; i++) {
		if(buf[i]=='\n') {
			std::cout << "Plan manager heard " << cmd
	                  << " from LISP" << endl;
			manager->runCommand(cmd);
			cmd.clear();
		} else
			cmd+=buf[i];
	}
  return 0;
}

void PlanManager::runAction(ACTIONS act) {
	actionArray[act]->DoStart();	
}

std::string PlanManager::printActionName(unsigned int action) {
	if(action == EFFECT_CHECK) return "EFFECT_CHECK";
	else return actionArray[action]->getName();
}

/*! @file plan_manager.cc
 * @brief implimentation of the PlanManager class
 * @author Daniel Casner (Creator)
 *
 */
